#include <stdio.h>
#include "oct.h"
#include "tr.h"
#include "st.h"
#include "my.h"

extern void exit();
extern int strcmp();
extern char *strcpy();

octStatus myLastStatus;

extern char *malloc();
extern char *realloc();
extern void free();

#define NIL(type)		((type *) 0)
#define ALLOC(type, num)	\
	((type *) malloc((unsigned)(num) * sizeof(type)))
#define REALLOC(type, obj, num)	\
	((type *) realloc((char *) obj,(unsigned) (num) * sizeof(type)))
#define FREE(object)		\
	(free((char *) object), object = 0)

#define strsav(s)		strcpy(ALLOC(char,strlen(s)+1), s)
#define min(a, b)		((a)<(b) ? (a) : (b))
#define max(a, b)		((a)>(b) ? (a) : (b))


/* LINTLIBRARY */

int myOctFailed(file, line, why, offending_object)
char *file;
int line;
char *why;
octObject *offending_object;
{
    (void) fprintf(stderr, 
	"Oct Assertion Failed (%d): source file %s at line %d\n", 
	myLastStatus, file, line);
    if (why != NIL(char)) {
	(void) fprintf(stderr, "%s", why);
	if (offending_object != NIL(octObject)) {
	    (void) fprintf(stderr, " %s", myFormatName(offending_object));
	}
	(void) fprintf(stderr, "\n");
    }
    octError(NIL(char));
    exit(1);
}


/* myOpen: obsolete */
octStatus myOpen(facet, cell_name, view_name, facet_name, mode)
octObject *facet;
char *cell_name, *view_name, *facet_name, *mode;
{
    return OpenFacet(facet, cell_name, view_name, facet_name, mode);
}


/* myOpenMaster: obsolete */
octStatus myOpenMaster(master, instance, facet, mode)
octObject *master, *instance;
char *facet, *mode;
{
    return OpenMaster(master, instance, facet, mode);
}


octStatus CreatePath2(c, t, p1x, p1y, p2x, p2y, width)
octObject *c, *t;
octCoord p1x, p1y, p2x, p2y;
octCoord width;
{
    struct octPoint p[2];
    octStatus status;

    t->type = OCT_PATH;
    t->contents.path.width = width;
    status = octCreate(c, t);
    if (status < OCT_OK) return status;

    p[0].x = p1x;
    p[0].y = p1y;
    p[1].x = p2x;
    p[1].y = p2y;
    status = octPutPoints(t, 2, p);
    return status;
}

/*
 *  myCopyContents: copy the contents facet of a cell
 */
octStatus myCopyContents(old_cell, old_view, new_cell, new_view)
char *old_cell;
char *old_view;
char *new_cell;
char *new_view;
{
    octStatus status;
    octObject newfacet, facet;

    newfacet.type = OCT_FACET;
    newfacet.contents.facet.cell = new_cell;
    newfacet.contents.facet.view = new_view;
    newfacet.contents.facet.facet = "contents";
    newfacet.contents.facet.version = OCT_CURRENT_VERSION;
    newfacet.contents.facet.mode = "w";

    facet.type = OCT_FACET;
    facet.contents.facet.cell = old_cell;
    facet.contents.facet.view = old_view;
    facet.contents.facet.facet = "contents";
    facet.contents.facet.version = OCT_CURRENT_VERSION;
    facet.contents.facet.mode = "r";
    status = octCopyFacet(&newfacet, &facet);
    return status;
}

/*
 *  myTerminalBB: Find the bounding box of the implementation of a terminal
 */

octStatus myTerminalBB(term, bb)
octObject *term;
struct octBox *bb;
{
    if (term->contents.term.instanceId == oct_null_id) {
	return myFormalTerminalBB(term, bb);
    } else {
	return myActualTerminalBB(term, bb);
    }
}



/*
 *  myActualTerminalBB: Find the bounding box of an actual terminal
 */

octStatus myActualTerminalBB(aterm, bb)
octObject *aterm;
struct octBox *bb;
{
    octObject instance, master, fterm, box;
    octStatus status;
    struct octTransform *transform;
    octCoord t;

    /* Get the instance from the terminal */
    status = GetById(&instance, aterm->contents.term.instanceId);
    if (status != OCT_OK) return status;

    /* open the interface facet of the master of the instance */
    status = OpenMaster(&master, &instance, "interface", "r");
    if (status < OCT_OK) return status;

    /* Find formal terminal (which matches this actual terminal) */
    status = GetByTermName(&master, &fterm, aterm->contents.term.name);
    if (status != OCT_OK) return status;

    /* Get the first 'terminal frame' (which must be a box) */
    status = myGenFirst(&fterm, OCT_BOX_MASK, &box);
    if (status != OCT_OK) return status;
    *bb = box.contents.box;

    /* Transform the box */
    transform = &instance.contents.instance.transform;
    tr_oct_transform(transform, &(bb->lowerLeft.x), &(bb->lowerLeft.y));
    tr_oct_transform(transform, &(bb->upperRight.x), &(bb->upperRight.y));
    if (bb->lowerLeft.x > bb->upperRight.x) {
	t = bb->lowerLeft.x;
	bb->lowerLeft.x = bb->upperRight.x;
	bb->upperRight.x = t;
    }
    if (bb->lowerLeft.y > bb->upperRight.y) {
	t = bb->lowerLeft.y;
	bb->lowerLeft.y = bb->upperRight.y;
	bb->upperRight.y = t;
    }
    return OCT_OK;
}


/*
 *  myFormalTerminalBB: Find the bounding box of a formal terminal
 */

octStatus myFormalTerminalBB(fterm, bb)
octObject *fterm;
struct octBox *bb;
{
    octObject aterm;
    octStatus status;

    /* Find implementing actual terminal */
    status = myGenFirst(fterm, OCT_TERM_MASK, &aterm);
    if (status != OCT_OK) return status;

    /* Return its bounding box */
    status = myActualTerminalBB(&aterm, bb);
    return status;
}

/*
 *  myGetByNameContainer: find the container of an object by name
 *
 *  Returns:
 *	OCT_OK if object was found
 *	OCT_NOT_FOUND if container was not found
 */

octStatus myGetByNameContainer(contents, obj, type, container_name)
octObject *contents;
octObject *obj;
int type;
char *container_name;
{
    octObjectMask mask;
    octStatus status;
    octGenerator gen;
    char *name;

    /* Start a generator on the containers */
    mask = 1 << type;
    status = octInitGenContainers(contents, mask, &gen);
    if (status != OCT_OK) return status;

    while ((status = octGenerate(&gen, obj)) == OCT_OK) {
	if ((name = myGetName(obj)) == NIL(char)) {
	    status = OCT_ERROR;
/*	    oct_error("Cannot distinguish object by name");*/
	} else if (strcmp(name, container_name) == 0) {
	    return OCT_OK;
	}
    }
    if (status != OCT_GEN_DONE) return status;
    return OCT_NOT_FOUND;
}

/*
 *  myGetName: return the name field of an object, or NIL(char) if the
 *  object has no name.
 */
char *myGetName(object)
octObject *object;
{
    char *name;

    switch (object->type) {
    case OCT_TERM: 
	name = object->contents.term.name; 
	break;
    case OCT_NET: 
	name = object->contents.net.name; 
	break;
    case OCT_INSTANCE: 
	name = object->contents.instance.name; 
	break;
    case OCT_PROP: 
	name = object->contents.prop.name; 
	break;
    case OCT_LAYER: 
	name = object->contents.layer.name; 
	break;
    case OCT_BAG: 
	name = object->contents.bag.name; 
	break;
    case OCT_LABEL: 
	name = object->contents.label.label; 
	break;
    default:
	name = NIL(char);
	break;
    }
    return name;
}


/*
 *  myPutName -- set the name field of an object to 'name'.  Returns 
 *  OCT_ERROR if the object does not have a name field.
 */
octStatus myPutName(object, name)
octObject *object;
char *name;
{
    switch (object->type) {
    case OCT_TERM: 
	object->contents.term.name = name; 
	break;
    case OCT_NET: 
	object->contents.net.name = name; 
	break;
    case OCT_INSTANCE: 
	object->contents.instance.name = name; 
	break;
    case OCT_PROP: 
	object->contents.prop.name = name; 
	break;
    case OCT_LAYER: 
	object->contents.layer.name = name; 
	break;
    case OCT_BAG: 
	object->contents.bag.name = name; 
	break;
    case OCT_LABEL: 
	object->contents.label.label = name; 
	break;
    default:
	return OCT_ERROR;
    }
    return OCT_OK;
}


/*
 * myObjectName -- return the name of an Oct type
 */
char *myObjectName(type)
int type;
{
    switch(type) {
    case OCT_BAG: 
	return "OCT_BAG";
    case OCT_BOX: 
	return "OCT_BOX";
    case OCT_CIRCLE: 
	return "OCT_CIRCLE";
    case OCT_EDGE: 
	return "OCT_EDGE";
    case OCT_FACET: 
	return "OCT_FACET";
    case OCT_INSTANCE: 
	return "OCT_INSTANCE";
    case OCT_LABEL: 
	return "OCT_LABEL";
    case OCT_LAYER: 
	return "OCT_LAYER";
    case OCT_NET: 
	return "OCT_NET";
    case OCT_POINT: 
	return "OCT_POINT";
    case OCT_POLYGON: 
	return "OCT_POLYGON";
    case OCT_PROP: 
	return "OCT_PROP";
    case OCT_TERM: 
	return "OCT_TERM";
    default:
	return "OCT_VERY_CONFUSED";
    }
}

/*
 *  myFormatName -- provide a reasonable 'name' description of an
 *  object, suitable for error messages.
 */
char *myFormatName(object)
octObject *object;
{
    static char s[1024];
    octStatus status;
    octObject inst;

    switch(object->type) {
	case OCT_BAG:
	case OCT_LABEL:
	case OCT_LAYER:
	case OCT_NET:
	case OCT_PROP:
	    (void) sprintf(s, "%s %s", 
		myObjectName(object->type), myGetName(object));
	    break;

	case OCT_TERM:
	    if (object->contents.term.instanceId == oct_null_id) {
		(void) sprintf(s, "formal terminal \"%s\"",
		    object->contents.term.name);
	    } else {
		status = GetById(&inst, object->contents.term.instanceId);
		(void) sprintf(s, "actual terminal \"%s\" of instance \"%s\"",
		    object->contents.term.name, 
		    status==OCT_OK ? inst.contents.instance.name : "--ERROR--");
	    }
	    break;
		    
	case OCT_INSTANCE:
	    (void) sprintf(s, "instance %s of %s:%s:%s:%s",
		object->contents.instance.name, 
		object->contents.instance.master, 
		object->contents.instance.view, 
		"interface",
		object->contents.instance.version);
	    break;
	
	case OCT_FACET:
	    (void) sprintf(s, "%s:%s:%s:%s",
		object->contents.facet.cell, object->contents.facet.view,
		object->contents.facet.facet, object->contents.facet.version);
	    break;

	default:
	    (void) strcpy(s, myObjectName(object->type));
	    break;
    }
    return s;
}

octStatus myModifyInstance(instance, new_cell, new_view)
octObject *instance;
char *new_cell;
char *new_view;
{
    octObject new_inst, new_term, term, facet;
    octGenerator gen;
    octStatus s;

    octGetFacet(instance, &facet);
    new_inst = *instance;
    new_inst.contents.instance.master = new_cell;
    new_inst.contents.instance.view = new_view;
    if ((s = octCreate(&facet, &new_inst)) != OCT_OK) return s;

    /*
     *  Copy attachements for objects attached to the instance
     */
    s = myAttachContents(instance, &new_inst, OCT_ALL_MASK &~ OCT_TERM_MASK);
    if (s != OCT_OK) return s; 
    s = myAttachContainers(instance, &new_inst, OCT_ALL_MASK &~ OCT_FACET_MASK);
    if (s != OCT_OK) return s; 

    /*
     *  Copy all objects attached to each terminal
     */
    if ((s=octInitGenContents(instance,OCT_TERM_MASK,&gen)) != OCT_OK) return s;
    while ((s = octGenerate(&gen, &term)) == OCT_OK) {

	/* Find corresponding terminal on the new cell */
	new_term = term;
	s = octGetByName(&new_inst, &new_term);
	if (s == OCT_OK) {
	    s = myAttachContents(&term, &new_term, OCT_ALL_MASK);
	    if (s != OCT_OK) return s; 
	    s = myAttachContainers(&term, &new_term, OCT_ALL_MASK);
	    if (s != OCT_OK) return s; 
	} else if (s == OCT_NOT_FOUND) {
	} else {
	    return s;
	}
    }
    if (s != OCT_GEN_DONE) return s;

    s = octDelete(instance);		/* delete the old instance */
    *instance = new_inst;		/* return the new instance */
    return s;
}


octStatus myAttachContents(src, dst, mask)
octObject *src;
octObject *dst;
octObjectMask mask;
{
    octGenerator gen;
    octObject obj;
    octStatus s;

    if ((s = octInitGenContents(src, mask, &gen)) != OCT_OK) return s;
    while ((s = octGenerate(&gen, &obj)) == OCT_OK) {
	if ((s = octAttach(dst, &obj)) != OCT_OK) return s;
    }
    if (s == OCT_GEN_DONE) s = OCT_OK;
    return s;
}


octStatus myAttachContainers(src, dst, mask)
octObject *src;
octObject *dst;
octObjectMask mask;
{
    octGenerator gen;
    octObject obj;
    octStatus s;

    if ((s = octInitGenContainers(src, mask, &gen)) != OCT_OK) return s;
    while ((s = octGenerate(&gen, &obj)) == OCT_OK) {
	if ((s = octAttach(&obj, dst)) != OCT_OK) return s;
    }
    if (s == OCT_GEN_DONE) s = OCT_OK;
    return s;
}


octStatus myDeleteContents(container, mask)
octObject *container;
octObjectMask mask;
{
    octGenerator gen;
    octObject obj;
    octStatus status;

    CHECK(octInitGenContents(container, mask, &gen));
    while ((status = octGenerate(&gen, &obj)) == OCT_OK) {
	CHECK(octDelete(&obj));
    }
    CHECK(status);
    return OCT_OK;
}

octStatus myFormalTerminalNet(fterm, net)
octObject *fterm;
octObject *net;
{
    return myGenFirstContainer(fterm, OCT_NET_MASK, net);
}


octStatus myFindFormal(fterm, aterm)
octObject *fterm, *aterm;
{
    octObject inst, master;
    octStatus status;

    status = GetById(&inst, aterm->contents.term.instanceId);
    if (status != OCT_OK) return status;
    status = OpenMaster(&master, &inst, "interface", "r");
    if (status < OCT_OK) return status;
    status = GetByTermName(&master, fterm, aterm->contents.term.name);
    return status;
}


octStatus myCountContents(object, mask, n)
octObject *object;
octObjectMask mask;
int *n;
{
    octStatus status;
    octObject obj;
    octGenerator gen;

    *n = 0;
    status = octInitGenContents(object, mask, &gen);
    if (status != OCT_OK) return status;
    while ((status == octGenerate(&gen, &obj)) == OCT_OK) {
	(*n)++;
    }
    if (status != OCT_OK) return status;
    return OCT_OK;
}


octStatus myContentsToArray(object, mask, array, array_n)
octObject *object;
octObjectMask mask;
octId **array;
int *array_n;
{
    octGenerator gen;
    octObject obj;
    octStatus s;
    int size;

    if ((s = octInitGenContents(object, mask, &gen)) != OCT_OK) return s;

    size = 3;
    if ((*array = ALLOC(octId, size)) == NIL(octId)) {
	return OCT_ERROR;
    }
    *array_n = 0;

    while ((s = octGenerate(&gen, &obj)) == OCT_OK) {
	if (*array_n >= size) {
	    size *= 2;
	    if ((*array = REALLOC(octId, *array, size)) == NIL(octId)) {
		return OCT_ERROR;
	    }
	}
	(*array)[*array_n] = obj.objectId;
	(*array_n)++;
    }
    if (s == OCT_GEN_DONE) s = OCT_OK;
    return s;
}

/*
 *  myGetPoints -- get the points from a PATH or POLYGON
 *
 *  A pointer to a static structure is returned; please do not free it
 */

octStatus myGetPoints(obj, num, arr)
octObject *obj;
int *num;
struct octPoint **arr;
{
    int status;
    int real_num_points;
    static int num_points = 0;
    static struct octPoint *points;

    if (num_points == 0) {
	num_points = 50;
	if ((points = ALLOC(struct octPoint, num_points)) == 0) {
	    return OCT_ERROR;
	}
    }
    
    real_num_points = num_points;
    status = octGetPoints(obj, &real_num_points, points);
    if (status != OCT_OK) {
	if (status != OCT_TOO_SMALL) return status;
	num_points = real_num_points;
	if ((points = REALLOC(struct octPoint, points, num_points)) == 0) {
	    return OCT_ERROR;
	}
	real_num_points = num_points;
	status = octGetPoints(obj, &real_num_points, points);
	if (status != OCT_OK) return status;
    }
    *num = real_num_points;
    *arr = points;
    return OCT_OK;
}

myPrintObject(object, indent)
octObject *object;
{
    octObject inst, layer;
    struct octBox bb;
    octId id;
    int i;

    for(i = 0; i < indent; i++) putchar(' ');

    switch (object->type) {
	case OCT_FACET:
	    (void) printf("*** OCT_FACET:\n");
	    (void) printf("cell=%s\n", object->contents.facet.cell);
	    (void) printf("view=%s\n", object->contents.facet.view);
	    (void) printf("facet=%s\n", object->contents.facet.facet);
	    (void) printf("version=%s\n", object->contents.facet.version);
	    (void) printf("mode=%s\n", object->contents.facet.mode);
	    break;
	case OCT_INSTANCE:
	    (void) printf("*** OCT_INSTANCE:\n");
	    (void) printf("name=%s\n", object->contents.instance.name);
	    (void) printf("master=%s\n", object->contents.instance.master);
	    (void) printf("view=%s\n", object->contents.instance.view);
	    (void) printf("version=%s\n", object->contents.instance.version);
	    if (octBB(object, &bb) >= OCT_OK) {
		(void) printf("bb= (%d,%d) (%d,%d)\n", 
		    bb.lowerLeft.x, bb.lowerLeft.y, 
		    bb.upperRight.x, bb.upperRight.y);
	    }
	case OCT_CIRCLE:
	    (void) printf("*** OCT_CIRCLE:\n");
	    (void) printf("startingAngle=%d endingAngle=%d\n",
		object->contents.circle.startingAngle,
		object->contents.circle.endingAngle);
	    (void) printf("innerRadius=%d outerRadius=%d\n",
		object->contents.circle.innerRadius,
		object->contents.circle.outerRadius);
	    (void) printf("center at (%d,%d)\n",
		object->contents.circle.center.x,
		object->contents.circle.center.y);
	    if (octBB(object, &bb) >= OCT_OK) {
		(void) printf("bb= (%d,%d) (%d,%d)\n", 
		    bb.lowerLeft.x, bb.lowerLeft.y, 
		    bb.upperRight.x, bb.upperRight.y);
	    }
	    if (myGenFirstContainer(object, OCT_LAYER_MASK, &layer) >= OCT_OK) 
		(void) printf( "on layer %s\n", layer.contents.layer.name);
	    (void) printf("\n");
	    break;
	case OCT_BOX:
	    bb = object->contents.box;
	    (void) printf("*** OCT_BOX: (%d,%d) (%d,%d)",
		bb.lowerLeft.x, bb.lowerLeft.y, 
		bb.upperRight.x, bb.upperRight.y);
	    if (myGenFirstContainer(object, OCT_LAYER_MASK, &layer) >= OCT_OK) 
		(void) printf( " on layer %s", layer.contents.layer.name);
	    (void) printf("\n");
	    break;
	case OCT_LABEL:
	    (void) printf("*** OCT_LABEL: label=%s", 
		object->contents.label.label);
	    if (myGenFirstContainer(object, OCT_LAYER_MASK, &layer) >= OCT_OK) 
		(void) printf( " on layer %s", layer.contents.layer.name);
	    (void) printf("\n");
	    break;
	case OCT_LAYER:
	    (void) printf("*** OCT_LAYER: name=%s\n", object->contents.layer.name);
	    break;
	case OCT_BAG:
	    (void) printf("*** OCT_BAG: name=%s\n", object->contents.bag.name);
	    break;
	case OCT_NET:
	    (void) printf("*** OCT_NET:  name=%s\n", object->contents.net.name);
	    break;
	case OCT_TERM:
	    (void) printf("*** OCT_TERM:  name=%s", object->contents.term.name);
	    id = object->contents.term.instanceId;
	    if (id == oct_null_id) {
		(void) printf(" (formal terminal)\n");
	    } else {
		if (GetById(&inst, object->contents.term.instanceId) >= OCT_OK) 
		    (void) printf( "  on instance %s\n", inst.contents.instance.name);
	    }
	    break;
	case OCT_PROP:
	    (void) printf("*** OCT_PROP: type=");
	    switch(object->contents.prop.type) {
		case OCT_NULL:
		    (void) printf("OCT_NULL\n");
		    break;
		case OCT_INTEGER:
		    (void) printf("OCT_INTEGER value=%d\n", 
			object->contents.prop.value.integer);
		    break;
		case OCT_REAL:
		    (void) printf("OCT_REAL value=%f\n", 
			object->contents.prop.value.real);
		    break;
		case OCT_STRING:
		    (void) printf("OCT_STRING value=%s\n", 
			object->contents.prop.value.string);
		    break;
		case OCT_ID:
		    (void) printf("OCT_ID\n");
		    break;
		case OCT_STRANGER:
		    (void) printf("OCT_STRANGER\n");
		    break;
		default:
		    (void) printf("+++UNKNOWN PROP TYPE+++\n");
		    break;
	    }
	    break;
	default:
	    (void) printf("+++UNKNOWN OBJECT TYPE+++\n");
	    break;
    }
}

octStatus myPrintObjects(container, mask)
octObject *container;
octObjectMask mask;
{
    octGenerator gen;
    octStatus status;
    octObject object;

    myPrintObject(container, 0);

    status = octInitGenContents(container, mask, &gen);
    if (status < OCT_OK) return status;
    while ((status = octGenerate(&gen, &object)) == OCT_OK) {
	myPrintObject(&object, 4);
    }
    if (status != OCT_GEN_DONE) return status;
    return OCT_OK;
}

octStatus myPrintObjects2(container, mask1, mask2)
octObject *container;
octObjectMask mask1, mask2;
{
    octGenerator gen, gen1;
    octStatus status;
    octObject object, object1;

    myPrintObject(container, 0);

    status = octInitGenContents(container, mask1, &gen);
    if (status < OCT_OK) return status;
    while ((status = octGenerate(&gen, &object)) == OCT_OK) {
	myPrintObject(&object, 4);
	status = octInitGenContents(&object, mask2, &gen1);
	if (status < OCT_OK) return status;
	while ((status = octGenerate(&gen1, &object1)) == OCT_OK) {
	    myPrintObject(&object1, 8);
	}
	if (status != OCT_GEN_DONE) return status;
    }
    if (status != OCT_GEN_DONE) return status;
    return OCT_OK;
}

octStatus myTransformContents(container, mask, transform)
octObject *container;
octObjectMask mask;
struct octTransform *transform;
{
    octObject geo;
    octGenerator gen;
    octStatus status;

    status = octInitGenContents(container, mask, &gen);
    if (status != OCT_OK) return status;
    while ((status = octGenerate(&gen, &geo)) == OCT_OK){
	status = octTransformAndModifyGeo(&geo, transform);
	if (status != OCT_OK) return status;
    }
    if (status == OCT_GEN_DONE) status = OCT_OK;
    return status;
}


octStatus myTransformFacet(facet, transform)
octObject *facet;
struct octTransform *transform;
{
    octObject layer;
    octGenerator gen;
    octStatus status;

    status = octInitGenContents(facet, OCT_LAYER_MASK, &gen);
    if (status != OCT_OK) return status;
    while ((status = octGenerate(&gen, &layer)) == OCT_OK) {
	status = myTransformContents(&layer, OCT_GEO_MASK, transform);
	if (status != OCT_OK) return status;
    }
    if (status != OCT_GEN_DONE) return status;

/*
 *  should probably multiply in the new transform
 *
    status = myScaleContents(facet, OCT_INSTANCE_MASK, scale);
    if (status != OCT_GEN_DONE) return status;
 */
    return OCT_OK;
}

octStatus myScaleContents(container, mask, scale)
octObject *container;
octObjectMask mask;
double scale;
{
    octObject geo;
    octGenerator gen;
    octStatus status;

    status = octInitGenContents(container, mask, &gen);
    if (status != OCT_OK) return status;
    while ((status = octGenerate(&gen, &geo)) == OCT_OK) {
	status = octScaleAndModifyGeo(&geo, scale);
	if (status != OCT_OK) return status;
    }
    if (status != OCT_GEN_DONE) return status;
    return OCT_OK;
}


octStatus myScaleFacet(facet, scale)
octObject *facet;
double scale;
{
    octObject layer;
    octGenerator gen;
    octStatus status;

    status = octInitGenContents(facet, OCT_LAYER_MASK, &gen);
    if (status != OCT_OK) return status;
    while ((status = octGenerate(&gen, &layer)) == OCT_OK) {
	status = myScaleContents(&layer, OCT_GEO_MASK, scale);
	if (status != OCT_OK) return status;
    }
    if (status != OCT_GEN_DONE) return status;

    status = myScaleContents(facet, OCT_INSTANCE_MASK, scale);
    if (status != OCT_GEN_DONE) return status;
    return OCT_OK;
}

struct sort_obj_struct {
    octObject obj;
    struct sort_obj_struct *next;
};

#define TYPE		struct sort_obj_struct
#define NEXT		next
#define FIELD		obj
#define SORT		linked_list_sort_helper
#define SORT1		linked_list_sort

/*
 *  'Generic' linked-list sorting 'package'
 *
 *  Use:
 *	#define TYPE		the linked-list type (usually a struct)
 *	#define NEXT		'next' field name in the linked-list structure
 *	#define SORT		sorting routine (see below)
 *
 *  Optional:
 *	#define DECL_SORT	'static' or undefined
 *	#define DECL_SORT1	'static' or undefined
 *	#define SORT1		sorting routine (see below)
 *	#define FIELD		select a subfield of the structure for the
 *				   compare function (default is to pass
 *				   a pointer to the structure)
 *	#include "lsort.h"
 *
 *  This defines up to two routines:
 *	SORT1(TYPE *list, int (*compare)(TYPE *x, TYPE *y))
 *	    sort the linked list 'list' according to the compare function
 *	    'compare'
 *
 *	SORT(TYPE *list, int (*compare)(TYPE *x, TYPE *y), int length)
 *	    sort the linked list 'list' according to the compare function
 *	    'compare'.  length is the length of the linked list.
 *
 *  Both routines gracefully handle length == 0 (in which case, list == 0 
 *  is also allowed).
 *
 *  By default, both routines are declared 'static'.  This can be changed
 *  using '#define DECL_SORT' or '#define DECL_SORT1'.
 *
 *  If field is used, then a pointer to the particular field is passed
 *  to the comparison function (rather than a TYPE *).
 */


#ifndef DECL_SORT1
#define DECL_SORT1 static
#endif

#ifndef DECL_SORT
#define DECL_SORT static
#endif

DECL_SORT TYPE *SORT();


#ifdef SORT1

DECL_SORT1 TYPE *SORT1(list_in, compare)
TYPE *list_in;
int (*compare)();
{
    register long cnt;
    register TYPE *p;

    /* Find the length of the list */
    for(p = list_in, cnt = 0; p != 0; p = p->NEXT, cnt++)
	;
    return SORT(list_in, compare, cnt);
}

#endif

DECL_SORT TYPE *SORT(list_in, compare, cnt)
TYPE *list_in;
int (*compare)();
long cnt;
{
    register TYPE *p, **plast, *list1, *list2;
    register long i;

    if (cnt > 1) {
	/* break the list in half */
	for(p = list_in, i = cnt/2-1; i > 0; p = p->NEXT, i--)
	    ;
	list1 = list_in;
	list2 = p->NEXT;
	p->NEXT = 0;

	/* Recursively sort the sub-lists (unless only 1 element) */
	if ((i = cnt/2) > 1) {
	    list1 = SORT(list1, compare, i);
	}
	if ((i = cnt - i) > 1) {
	    list2 = SORT(list2, compare, i);
	}

	/* Merge the two sorted sub-lists */
	plast = &list_in;
	for(;;) {
#ifdef FIELD
	    if ((*compare)(&list1->FIELD, &list2->FIELD) <= 0) {
#else
	    if ((*compare)(list1, list2) <= 0) {
#endif
		*plast = list1;
		plast = &(list1->NEXT);
		if ((list1 = list1->NEXT) == 0) {
		    *plast = list2;
		    break;
		}
	    } else {
		*plast = list2;
		plast = &(list2->NEXT);
		if ((list2 = list2->NEXT) == 0) {
		    *plast = list1;
		    break;
		}
	    }
	}
    }

    return list_in;
}

octStatus mySortContents(container, mask, compare)
octObject *container;
octObjectMask mask;
int (*compare)();
{
    octGenerator gen;
    octObject obj;
    octStatus status;
    struct sort_obj_struct *linknext, *head, *link;

    /* Gen all the objects into a linked list */
    head = NIL(struct sort_obj_struct);
    status = octInitGenContents(container, mask, &gen);
    if (status != OCT_OK) return status;
    while ((status = octGenerate(&gen, &obj)) == OCT_OK) {
	if ((link = ALLOC(struct sort_obj_struct, 1)) == 0) {
	    return OCT_ERROR;
	}
	link->next = head;
	head = link;
	link->obj = obj;

	status = octDetach(container, &obj);
	if (status != OCT_OK) return status;
    }
    if (status != OCT_GEN_DONE) return status;

    /* Sort this linked list */
    head = linked_list_sort(head, compare);

    /* Re-attach the objects to the container */
    for(link = head; link != 0; link = linknext) {
       status = octAttach(container, &link->obj);
       if (status != OCT_OK) return status;
       linknext = link->next;
       FREE(link);
    }

    return OCT_OK;
}

octStatus myUniqNames(container, mask)
octObject *container;
octObjectMask mask;
{
    octObject obj;
    octGenerator gen;
    octStatus status;
    char *name, dummyname[1024];
    st_table *name_table;
    int cnt, count;

    name_table = st_init_table(strcmp, st_strhash);
    cnt = 0;

    CHECK(octInitGenContents(container, mask, &gen));
    while ((status = octGenerate(&gen, &obj)) == OCT_OK) {
	name = myGetName(&obj);
	if (name == NIL(char)) return OCT_ERROR;
	if (name[0] == '\0') {
	    /* make sure the name is not empty */
	    (void) sprintf(dummyname, "%s-%d", myObjectName(obj.type), cnt++);
	    CHECK(myPutName(&obj, dummyname));
	    CHECK(octModify(&obj));
	    CHECK(GetById(&obj, obj.objectId));
	    name = myGetName(&obj);
	}
	if (st_lookup(name_table, name, NIL(char *))) {
	    /* loop until we can generate a unique name */
	    count = 2;
	    do {
		(void) sprintf(dummyname, "%s-{%d}", name, count++);
	    } while (st_lookup(name_table, dummyname, NIL(char *)));

	    /* change the objects name */
	    CHECK(myPutName(&obj, dummyname));
	    CHECK(octModify(&obj));
	    CHECK(GetById(&obj, obj.objectId));
	    name = myGetName(&obj);
	}

	(void) st_insert(name_table, name, NIL(char));
    }
    CHECK(status);
    (void) st_free_table(name_table);
    return OCT_OK;
}
